"""
get_cfs.py
----------

Script which parses costfunctions.py to see what cost functions are available.
Used by the frontend.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import ast
from pathlib import Path


def main():
    """
    Parses costfunctions.py to get the top-level functions. Prints them to
    stdout with a space as the delimiter. This allows the frontend to read them
    in using Popen.communicate().
    """
    cf_file = Path(__file__).resolve().parent / "costfunctions.py"
    with open(cf_file, 'r') as fp:
        tree = ast.parse(fp.read())
    functions = [func.name for func in tree.body
                 if isinstance(func, ast.FunctionDef)]
    print(" ".join(functions))


if __name__ == "__main__":
    main()



"""
SPDX-License-Identifier: GPL-3.0-or-later
"""

import json
import re
from pathlib import Path

import numpy as np
import pandas as pd


def parse_log(fname):
    """
    Parse a poise.log file.

    Parameters
    ----------
    fname : str or pathlib.Path
        Path to poise.log file, or the folder containing it (this would be the
        TopSpin EXPNO folder).

    Returns
    -------
    log_df : pandas.DataFrame
        Dataframe with rows corresponding to optimisations which successfully
        terminated. The time taken is given in seconds.
    """
    # Check if the file exists
    fname = Path(fname)
    if fname.is_dir():
        fname = fname / "poise.log"
    if not fname.exists():
        raise FileNotFoundError(f"The log file '{fname}' was not found.")

    empty_run = {"routine": None, "init": None, "cf": None, "algo": None,
                 "au": None, "lb": None, "ub": None, "tol": None,
                 "xbest": None, "feval": None, "elapsed": None,
                 }
    current_run = dict(empty_run)  # make a copy
    all_runs = []   # List of all successful optimisation runs

    def parse_list_params(line):
        """
        Extracts init, lb, ub, tol from the line.
        """
        # Get the second half of the line
        rhs = line.split("-", maxsplit=1)[1].strip()
        # Convert '[4.]' to '[4]', otherwise json.loads() trips up
        rhs = re.sub(r"\.([^\d])", r"\1", rhs)
        # Evaluate the string and remove the list if it's a singleton
        rhs = json.loads(rhs)
        return rhs[0] if len(rhs) == 1 else rhs

    with open(fname, "r") as fp:
        for line in fp:
            # Check if there is a completed optimisation stored.
            # If so, add it to all_runs then reset the current run.
            if current_run["elapsed"] is not None:
                all_runs.append(current_run)
                current_run = dict(empty_run)
            # Start of a new optimisation, reset all values
            if "===================" in line:
                current_run = dict(empty_run)
            # Parse lines
            elif line.startswith("Routine name"):
                current_run["routine"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Optimisation parameters"):
                p = line.split("-", maxsplit=1)[1].strip().replace("'", '"')
                current_run["param"] = json.loads(p)
            elif line.startswith("Initial values"):
                current_run["init"] = parse_list_params(line)
            elif line.startswith("Lower bounds"):
                current_run["lb"] = parse_list_params(line)
            elif line.startswith("Upper bounds"):
                current_run["ub"] = parse_list_params(line)
            elif line.startswith("Tolerances"):
                current_run["tol"] = parse_list_params(line)
            elif line.startswith("Cost function  "):
                current_run["cf"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("AU programme"):
                current_run["au"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Optimisation algorithm"):
                current_run["algo"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Best values found"):
                current_run["xbest"] = parse_list_params(line)
            elif line.startswith("Cost function at"):
                fbest = line.split("-", maxsplit=1)[1].strip()
                current_run["fbest"] = float(fbest)
            elif line.startswith("Number of spectra ran"):
                feval = line.split("-", maxsplit=1)[1].strip()
                current_run["feval"] = int(feval)
            elif line.startswith("Total time taken"):
                elapsed = line.split("-", maxsplit=1)[1].strip()
                elapsed = [int(i) for i in elapsed.split(":")]
                current_run["elapsed"] = (3600*elapsed[0]
                                          + 60*elapsed[1]
                                          + elapsed[2])

    # Check one more time at the end of the file to make sure that the
    # last entry is recorded.
    if current_run["elapsed"] is not None:
        all_runs.append(current_run)
        current_run = dict(empty_run)

    # Construct dataframe and return.
    data = {}
    data["routine"] = [run["routine"] for run in all_runs]
    data["initial"] = [run["init"] for run in all_runs]
    data["param"] = [run["param"] for run in all_runs]
    data["lb"] = [run["lb"] for run in all_runs]
    data["ub"] = [run["ub"] for run in all_runs]
    data["tol"] = [run["tol"] for run in all_runs]
    data["algorithm"] = [run["algo"] for run in all_runs]
    data["costfn"] = [run["cf"] for run in all_runs]
    data["auprog"] = [run["au"] for run in all_runs]
    data["optimum"] = [run["xbest"] for run in all_runs]
    data["fbest"] = [run["fbest"] for run in all_runs]
    data["nfev"] = [run["feval"] for run in all_runs]
    data["time"] = [run["elapsed"] for run in all_runs]
    return pd.DataFrame(data=data)


def xmean(series):
    """
    An aggregator function that allows you to average lists in DataFrame
    columns. This is useful for analysing optimisation runs with multiple
    parameters.

    For example:

    >>> df = parse_log(logfile)
    >>> # works for 1D, but not for nD optimisations
    >>> df.groupby("algorithm").mean()
    >>> # works for all dimensions.
    >>> df.groupby("algorithm").agg(xmean)
    """
    return np.mean(np.array(series.to_list()), axis=0).tolist()



import numpy as np
import pytest

from nmrpoise.poise_backend.optpoise import (nelder_mead,
                                             multid_search,
                                             pybobyqa_interface,
                                             deco_count)
from nmrpoise.poise_backend.backend import scale, unscale


def test_deco_count():
    @deco_count
    def peep():
        print("PEEP")

    n = 10
    for i in range(n):
        peep()
    assert peep.calls == n


@deco_count
def rosenbrock(x, arg1=None, arg2=None):
    # Using scipy's definition. We have a couple of dummy arguments because
    # because PyBOBYQA *requires* optimargs and then passes them to the cost
    # function, so rosenbrock() has to accept a couple of extra arguments,
    # even though it doesn't do anything with them.
    return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)


@deco_count
def quadratic(x):
    return np.sum(x ** 2)


# This xtol is extremely large, but more closely parallels the
# real-life situations we are likely to face.
x0 = np.array([1.3, 0.7, 0.8, 1.9, 1.2])
xtol = np.array([1e-2] * len(x0))
nm_simplex_methods_nexpts = {
    "spendley": 1,
    "axis": 1,
    "random": 20,
}


def test_errors():
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                simplex_method="Penguin")
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0[:-1], xtol=xtol,
                                simplex_method="Penguin")
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=np.ones(len(x0)),
                                simplex_method="Penguin")


def test_NM_accuracy():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        quadratic.calls = 0  # reset fevals
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                simplex_method=method)
        if not np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    simplex_method=method)
        assert np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2)


def test_NM_iters():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    simplex_method=method)
            count += optResult.niter
        count /= nexpt
        assert count < 500


def test_NM_fevals():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    simplex_method=method)
            count += optResult.nfev
        count /= nexpt
        assert count < 850


mds_simplex_methods_nexpts = {
    "spendley": 1,
    "axis": 1,
}


def test_MDS_accuracy():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        quadratic.calls = 0  # reset fevals
        optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                  simplex_method=method)
        assert np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2)


def test_MDS_iters():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                      simplex_method=method)
            count += optResult.niter
        count /= nexpt
        assert count < 500


def test_MDS_fevals():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                      simplex_method=method)
            count += optResult.nfev
        count /= nexpt
        assert count < 1000


def test_maxfevals_reached():
    optResult = nelder_mead(cf=quadratic, x0=x0, xtol=([1e-6] * len(x0)),
                            maxfev=10)
    assert optResult.message == "Maximum function evaluations reached."



"""
dosy_opt.py
-----------

Script that optimises parameters for a diffusion experiment. Currently only
works with oneshot DOSY (Pelta et al., Magn. Reson. Chem. 2002, 40 (13),
S147–S152, DOI: 10.1002/mrc.1107).

This is meant to be called from the TopSpin command line on a 2D diffusion
experiment.

Note that this script does not tweak the gradient lengths ("little delta"), so
these should be set to a sensible value (e.g. 1 ms) before running this script.

The algorithm is as follows:

 1. Create 1D versions of the diffusion experiment, one with the minimum
    gradient (10%, the "reference spectrum", expno 99998) and one with the
    maximum gradient (80%, "optimisation spectrum", expno 99999).
 2. Acquires both spectra.
 3. Check whether the opt. spectrum has at least 75% signal attenuation
    relative to the ref. spectrum. If not, increases the diffusion delay
    ("big Delta") and repeats steps 1 and 2.
 4. If the attenuation is sufficient, it then runs a POISE optimisation on the
    opt. spectrum to find the gradient strength at which 75% signal attenuation
    is achieved.
 5. Once this gradient strength has been found, returns to the 2D experiment
    and runs the `dosy` Bruker AU programme using the new parameter.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import time
import os

# Find the correct parameter set for the 1D sequence.
if GETPAR("PULPROG") == "doneshot_2d_jy":
    pulprog = "doneshot_1d_jy"
else:
    MSG("dosy_opt: unsupported pulse programme")
# Other variants can be added here.

# Get the expno of the 2D dataset, and store the parameter set. This is
# important because we want the optimisation to inherit most parameters, e.g.
# D1, SW, etc. from the parent experiment (which has presumably been set up by
# the user).
original_2d_expno = int(CURDATA()[1])
original_2d_dataset = CURDATA()
XCMD("wpar dosytemp all")

# Set up the reference experiment.
reference_expno = 99998
reference_dataset = list(original_2d_dataset)
reference_dataset[1] = str(reference_expno)
NEWDATASET(reference_dataset, None, "dosytemp")
RE(reference_dataset)
PUTPAR("PULPROG", pulprog)
PUTPAR("PARMODE", "1D")
PUTPAR("PPARMOD", "1D")
PUTPAR("GPZ 1", "10")
XCMD("sendgui browse_update_tree")
# Acquire reference dataset.
XCMD("poise_1d", wait=WAIT_TILL_DONE)

# Set up optimisation expno.
optimisation_expno = 99999
optimisation_dataset = list(original_2d_dataset)   # make a copy
optimisation_dataset[1] = str(optimisation_expno)
NEWDATASET(optimisation_dataset, None, "dosytemp")
RE(optimisation_dataset)
PUTPAR("PULPROG", pulprog)
PUTPAR("PARMODE", "1D")
PUTPAR("PPARMOD", "1D")
PUTPAR("GPZ 1", "80")
XCMD("sendgui browse_update_tree")

# Before we can run POISE, we need to delete the acqu2s and proc2s files from
# both ref. and opt. datasets (they are there because are inherited from the 2D
# parameter set we set the experiments up with). If we don't do this, then
# POISE gets very confused as it looks at the file structure and thinks that
# they are 2D spectra.
ref_expnof = os.path.join(reference_dataset[3],
                          reference_dataset[0],
                          reference_dataset[1])
opt_expnof = os.path.join(optimisation_dataset[3],
                          optimisation_dataset[0],
                          optimisation_dataset[1])
for f in (os.path.join(ref_expnof, "acqu2s"),
          os.path.join(opt_expnof, "acqu2s"),
          os.path.join(ref_expnof, "pdata", reference_dataset[2], "proc2s"),
          os.path.join(opt_expnof, "pdata", optimisation_dataset[2], "proc2s")
          ):
    if os.path.isfile(f):
        os.remove(f)

while True:
    # Here we use the `--maxfev 1` trick to evaluate a cost function at 80%
    # gradient. The cost function is stored as the `TI` parameter after the
    # experiment has been recorded.
    RE(optimisation_dataset)
    XCMD("poise dosy1d_aux -a nm --maxfev 1 -q", wait=WAIT_TILL_DONE)
    cf = float(GETPAR("TI"))
    # The dosy1d_aux cost function is (opt. intensity/ref. intensity) - 0.25.
    # If the opt. spectrum has been attenuated too little, then this value will
    # be positive.
    if cf > 0:
        # Increase D20, i.e. big Delta, by 100 ms
        current_d20 = float(GETPAR("D 20"))
        new_d20 = current_d20 + 0.1
        PUTPAR("D 20", str(new_d20))
        # Set it in the reference spectrum too
        RE(reference_dataset)
        PUTPAR("D 20", str(new_d20))
        # Re-acquire the reference spectrum.
        XCMD("xau poise_1d", wait=WAIT_TILL_DONE)
    # If the cost function is negative, then that means it's sufficiently
    # attenuated, and we can go on to the actual optimisation.
    else:
        break

# At this point, we know that 80% gradient strength provides at least 75%
# attenuation. So we can run an optimisation to find the actual value that does
# provide 75% attenuation (which is somewhere between 10% and 80%).
RE(optimisation_dataset)
XCMD("poise dosy1d -a bobyqa -q", wait=WAIT_TILL_DONE)
# Save the final parameters.
best_d20 = GETPAR("D 20")
best_gpz1 = GETPAR("GPZ 1")

# Jump back to original 2D expt. Set D20, but not GPZ1 as that should always be
# 100% (the actual gradient amplitude is controlled by Difframp, which will be
# generated by the `dosy` AU programme).
RE(original_2d_dataset)
PUTPAR("D 20", best_d20)
PUTPAR("GPZ 1", "100")
# Generate Difframp, etc. and then run the AU programme.
XCMD("xau dosy 10 {} 16 l y".format(best_gpz1))

y = time.time()
MSG(str(y - x))



from pathlib import Path

import numpy as np

from nmrpoise.poise_backend import backend as be


def test_scaleby_bounds():
    rng = np.random.default_rng()
    # Generate and sort a random 3x20 matrix.
    mat = rng.uniform(low=-500, high=500, size=(3, 20))
    mat.sort(axis=0)
    lb, val, ub = mat[0], mat[1], mat[2]
    # Generate random percentage tolerances.
    tol = rng.uniform(low=0.01, high=0.5, size=(1, 20))
    # Convert to actual tolerances.
    tol = tol * (ub - lb)
    # Scale.
    sval, slb, sub, stol = be.scale(val, lb, ub, tol)
    # Make sure that lb's and ub's are zeros and ones.
    assert np.allclose(slb, np.zeros(np.shape(val)))
    assert np.allclose(sub, np.ones(np.shape(val)))
    # Make sure that unscaling gives the right things.
    usval = be.unscale(sval, lb, ub, tol)
    assert np.allclose(usval, val)


def test_scaleby_tols():
    rng = np.random.default_rng()
    # Generate and sort a random 3x20 matrix.
    mat = rng.uniform(low=-500, high=500, size=(3, 20))
    mat.sort(axis=0)
    lb, val, ub = mat[0], mat[1], mat[2]
    # Generate random percentage tolerances.
    tol = rng.uniform(low=0.01, high=0.5, size=(1, 20))
    # Convert to actual tolerances.
    tol = tol * (ub - lb)
    # Scale by tolerances.
    sval, slb, sub, stol = be.scale(val, lb, ub, tol, scaleby="tols")
    # Make sure that lb's and tol's are 0's and 0.03's.
    assert np.allclose(slb, np.zeros(np.shape(val)))
    assert np.allclose(stol, 0.03 * np.ones(np.shape(val)))
    # Check that unscaling gives back the same values.
    usval = be.unscale(sval, lb, ub, tol, scaleby="tols")
    assert np.allclose(usval, val)


def test_get_routine_cf():
    routine_id = "p1cal"   # serialised on AV600
    p_routine_dir = Path(__file__).parent / "test_data"

    routine, cf = be.get_routine_cf(routine_id, p_routine_dir)
    assert routine.name == "p1cal"
    assert routine.pars == ["p1"]
    assert routine.lb == [40]
    assert routine.ub == [56]
    assert routine.init == [48]
    assert routine.tol == [0.2]
    assert routine.cf == "minabsint"
    assert routine.au == "poise_1d"
